@@ -1,10 +1,11 @@
 // Copyright 2012-2013 Tinyarray authors.
 //
 // This file is part of Tinyarray.  It is subject to the license terms in the
-// LICENSE file found in the top-level directory of this distribution and at
-// http://git.kwant-project.org/tinyarray/about/LICENSE.  A list of Tinyarray
-// authors can be found in the README file at the top-level directory of this
-// distribution and at http://git.kwant-project.org/tinyarray/about/.
+// file LICENSE.rst found in the top-level directory of this distribution and
+// at https://gitlab.kwant-project.org/kwant/tinyarray/blob/master/LICENSE.rst.
+// A list of Tinyarray authors can be found in the README.rst file at the
+// top-level directory of this distribution and at
+// https://gitlab.kwant-project.org/kwant/tinyarray.
 
 #include <Python.h>
 #include "array.hh"
@@ -17,8 +18,12 @@ int dtype_converter(const PyObject *ob, Dtype *dtype)
 {
     if (ob == Py_None) {
         *dtype = default_dtype;
+    #if PY_MAJOR_VERSION < 3
     } else if (ob == (PyObject *)(&PyInt_Type) ||
                ob == (PyObject *)(&PyLong_Type)) {
+    #else
+    } else if (ob == (PyObject *)(&PyLong_Type)) {
+    #endif
         *dtype = LONG;
     } else if (ob == (PyObject *)(&PyFloat_Type)) {
         *dtype = DOUBLE;
@@ -122,7 +127,7 @@ PyObject *zeros(PyObject *, PyObject *args)
 }
 
 PyDoc_STRVAR(zeros_doc,
-"zeros(shape, dtype="DEFAULT_DTYPE")\n\n\
+"zeros(shape, dtype=" DEFAULT_DTYPE ")\n\n\
 Return an array of given shape and type, filled with zeros.");
 
 PyObject *ones(PyObject *, PyObject *args)
@@ -131,7 +136,7 @@ PyObject *ones(PyObject *, PyObject *args)
 }
 
 PyDoc_STRVAR(ones_doc,
-"ones(shape, dtype="DEFAULT_DTYPE")\n\n\
+"ones(shape, dtype=" DEFAULT_DTYPE ")\n\n\
 Return an array of given shape and type, filled with ones.");
 
 template <typename T>
@@ -170,7 +175,7 @@ PyObject *identity(PyObject *, PyObject *args)
 }
 
 PyDoc_STRVAR(identity_doc,
-"identity(n, dtype="DEFAULT_DTYPE")\n\n\
+"identity(n, dtype=" DEFAULT_DTYPE ")\n\n\
 Return an identity matrix of given size and dtype.");
 
 PyObject *array(PyObject *, PyObject *args)
@@ -194,7 +199,7 @@ PyObject *matrix(PyObject *, PyObject *args)
     Dtype dtype = NONE;
     if (!PyArg_ParseTuple(args, "O|O&", &src, dtype_converter, &dtype))
         return 0;
-    return matrix_from_arraylike(src, &dtype);
+    return array_from_arraylike(src, &dtype, Dtype(0), true);
 }
 
 PyDoc_STRVAR(matrix_doc,
@@ -213,7 +218,9 @@ PyObject *transpose(PyObject *, PyObject *args)
     Dtype dtype = NONE;
     a = array_from_arraylike(a, &dtype);
     if (!a) return 0;
-    return transpose_dtable[int(dtype)](a, 0);
+    PyObject *result = transpose_dtable[int(dtype)](a, 0);
+    Py_DECREF(a);
+    return result;
 }
 
 PyDoc_STRVAR(transpose_doc,
@@ -302,7 +309,11 @@ PyMethodDef functions[] = {
     {"add", binary_ufunc<Add>, METH_VARARGS, binary_ufunc_doc},
     {"subtract", binary_ufunc<Subtract>, METH_VARARGS, binary_ufunc_doc},
     {"multiply", binary_ufunc<Multiply>, METH_VARARGS, binary_ufunc_doc},
+#if PY_MAJOR_VERSION < 3
     {"divide", binary_ufunc<Divide>, METH_VARARGS, binary_ufunc_doc},
+#else
+    {"divide", binary_ufunc<True_divide>, METH_VARARGS, binary_ufunc_doc},
+#endif
     {"remainder", binary_ufunc<Remainder>, METH_VARARGS, binary_ufunc_doc},
     {"floor_divide", binary_ufunc<Floor_divide>, METH_VARARGS,
      binary_ufunc_doc},
